﻿using System;
using System.ComponentModel;
using System.IO.Ports;

namespace TI83PlusBootExec {

	[Description("Black Link")]
	internal class BlackLink : LinkCable, IDisposable {

		private SerialPort port;
		private bool disposedValue;



		public BlackLink(SerialPort port) : base() {

			this.port = port;


			if (!this.port.IsOpen) {
				this.port.Open();
			}

			PinD0 = true;
			PinD1 = true;

		}

		protected virtual void Dispose(bool disposing) {
			if (!disposedValue) {
				if (disposing) {
					this.port.Dispose();
				}
				disposedValue = true;
			}
		}

		public void Dispose() {
			Dispose(disposing: true);
			GC.SuppressFinalize(this);
		}

		public bool PinD0 {
			get {
				return port.CtsHolding;
			}
			set {
				port.RtsEnable = value;
			}
		}

		public bool PinD1 {
			get {
				return port.DsrHolding;
			}
			set {
				port.DtrEnable = value;
			}
		}

		private DateTime CreateTimeout() {
			var timeout = DateTime.Now;
			timeout += Timeout;
			return timeout;
		}

		public override bool TryReadByte(out byte value) {

			value = 0;
			
			// only try to receive data if one of D0/D1 have gone low
			if (this.PinD0 && this.PinD1) {
				return false;
			}

			// receive 8 bits
			for (int i = 0; i < 8; ++i) {

				var timeout = CreateTimeout();
				bool? bit = null;

				// which pin has gone low?
				while (DateTime.Now < timeout) {
					if (!this.PinD0 && this.PinD1) {
						this.PinD1 = false;
						bit = false;
						break;
					} else if (!this.PinD1 && this.PinD0) {
						this.PinD0 = false;
						bit = true;
						break;
					}
				}

				// did we get a bit?
				if (!bit.HasValue) {
					this.PinD0 = true;
					this.PinD1 = true;
					return false;
				}

				// shift it in
				value >>= 1;
				if (bit.Value) {
					value |= 0x80;
				}

				// wait for the other line to go back high
				timeout = CreateTimeout();
				while (DateTime.Now < timeout) {
					if (!bit.Value && this.PinD0) {
						this.PinD1 = true;
						bit = null;
						break;
					} else if (bit.Value && this.PinD1) {
						this.PinD0 = true;
						bit = null;
						break;
					}
				}

				this.PinD0 = true;
				this.PinD1 = true;

				// did the line go back high?
				if (bit.HasValue) {
					return false;
				}
			}

			return true;

		}

		public override bool TryWriteByte(byte value) {

			// only try to send data if both of D0/D1 are still high
			if (!this.PinD0 || !this.PinD1) {
				return false;
			}

			// send 8 bits
			for (int i = 0; i < 8; ++i) {

				var timeout = CreateTimeout();
				
				// fetch the bit to send
				bool? bit = (value & 1) != 0;
				value >>= 1;

				if (!bit.Value) {
					this.PinD0 = false;
				} else {
					this.PinD1 = false;
				}

				// wait for the bit to be acknowledged
				timeout = CreateTimeout();
				while (DateTime.Now < timeout) {
					if (!bit.Value && !this.PinD1) {
						this.PinD0 = true;
						bit = null;
						break;
					} else if (bit.Value && !this.PinD0) {
						this.PinD1 = true;
						bit = null;
						break;
					}
				}

				this.PinD0 = true;
				this.PinD1 = true;

				// did the bit get acknowledged?
				if (bit.HasValue) {
					return false;
				}

				// wait for the both lines to go back high
				timeout = CreateTimeout();
				while (DateTime.Now < timeout) {
					if (this.PinD0 && this.PinD1) {
						bit = true;
						break;
					}
				}

				if (!bit.HasValue) {
					return false;
				}


			}
			return true;

		}

	}
}
