﻿using System;
using System.ComponentModel;
using System.Globalization;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Text;

namespace TI83PlusBootExec {
	internal class Program {

		static void Main(string[] args) {

			// show the title
			Console.WriteLine("TI-83 Plus BootExec - Ben Ryves 2023 | Based on Overflow by Brandon Wilson 2009");
			Console.WriteLine("-------------------------------------------------------------------------------");

			if (args.Length == 0) {
				Console.WriteLine("Usage: bootexec [-flashunlock] -zeropad[=num] COM1 file.8xp");
				Console.WriteLine("   or: bootexec [-flashunlock] -fixed[=hex] COM1 file.8xp");
				Console.WriteLine("If using a SilverLink use 'USB' as port name and ensure WinUSB driver is installed.");
				return;
			}


			// name of the serial port and payload file
			string cablePortName = null;
			string payloadFileName = null;

			// will we unlock the flash?
			bool unlockFlash = false;

			// will we pad the packet with zeroes, and if so how many?
			bool zeroPadPacket = false;
			int zeroPadPacketCount = -1;

			// is it a fixed packet size, and if so where's the variable in the stack?
			bool fixedPacketSize = false;
			ushort fixedPacketSizeStackPointer = 0;

			// parse command-line arguments
			foreach (string arg in args) {

				if (arg.ToLowerInvariant().StartsWith("-flashunlock")) {
					unlockFlash = true;
				} else if (arg.ToLowerInvariant().StartsWith("-zeropad")) {
					zeroPadPacket = true;
					if (arg.Contains("=")) {
						if (int.TryParse(arg.Substring(arg.IndexOf("=") + 1), NumberStyles.Number, CultureInfo.InvariantCulture, out int val) && val >= 0) {
							zeroPadPacketCount = val;
						} else {
							Console.WriteLine("Could not parse number in argument '{0}'.", arg);
							return;
						}
					}
				} else if (arg.ToLowerInvariant().StartsWith("-fixed")) {
					fixedPacketSize = true;
					if (arg.Contains("=")) {
						if (int.TryParse(arg.Substring(arg.IndexOf("=") + 1), NumberStyles.HexNumber, CultureInfo.InvariantCulture, out int val)) {
							fixedPacketSizeStackPointer = (ushort)val;
						} else {
							Console.WriteLine("Could not parse hex number in argument '{0}'.", arg);
							return;
						}
					}
				} else if (Array.Exists(SerialPort.GetPortNames(), n => n.ToUpperInvariant() == arg.ToUpperInvariant())) {
					cablePortName = Array.Find(SerialPort.GetPortNames(), n => n.ToUpperInvariant() == arg.ToUpperInvariant());
				} else if (arg.ToUpperInvariant() == "USB") {
					cablePortName = arg.ToUpperInvariant();
				} else if (File.Exists(arg)) {
					payloadFileName = arg;
				} else {
					Console.WriteLine("Unrecognised argument '{0}'.", arg);
					return;
				}

			}

			// are the arguments sensible?
			if (zeroPadPacket && fixedPacketSize) {
				Console.WriteLine("Only use one of the -zeropad or -fixed options.");
				return;
			} else if (!zeroPadPacket && !fixedPacketSize) {
				Console.WriteLine("You must choose one of the -zeropad or -fixed options.");
				return;
			} else if (cablePortName == null) {
				Console.WriteLine("Cable port name not specified.");
				return;
			} else if (payloadFileName == null) {
				Console.WriteLine("Program file name not specified.");
				return;
			}

			// default stack pointers
			if (fixedPacketSize && fixedPacketSizeStackPointer < 0x8000) {
				if (unlockFlash) {
					fixedPacketSizeStackPointer = 0xFFBF;
				} else {
					fixedPacketSizeStackPointer = 0xFFC1;
				}

			}

			// try to open the payload
			byte[] payload;
			string payloadName;
			try {
				using (var f = File.OpenRead(payloadFileName))
				using (var br = new BinaryReader(f)) {
					// open the group file and check it contains a single variable
					var groupFile = new GroupFile(br);
					if (groupFile.Variables.Count != 1) {
						throw new InvalidDataException(string.Format("Group file contains {0} variables.", groupFile.Variables.Count));
					}
					// check the variable has at least two bytes of data in it
					var variable = groupFile.Variables[0];
					payloadName = Encoding.ASCII.GetString(variable.Name).Split('\0')[0];
					if (variable.Data.Length < 2) {
						throw new InvalidDataException(string.Format("Variable {0} is too small.", payloadName));
					}
					// check for the AsmPrgm token
					if (variable.Data[0] != 0xBB || variable.Data[1] != 0x6D) {
						throw new InvalidDataException(string.Format("Variable {0} does not start with AsmPrgm token.", payloadName));
					}
					// copy the payload program data, minus the AsmPrgm token
					payload = new byte[variable.Data.Length - 2];
					Array.Copy(variable.Data, 2, payload, 0, payload.Length);
				}
			} catch (Exception ex) {
				Console.WriteLine("Could not open the program file: {0}", ex.Message);
				return;
			}
			if (payload == null) {
				Console.WriteLine("No program payload to send.");
				return;
			}

			// link cable
			LinkCable linkPort = null;

			if (cablePortName == "USB") {
				
				// try to open the USB port using TI's driver
				foreach (var item in SilverLinkTI.GetCables()) {
					try {
						item.Open();
						linkPort = item;
						break;
					} catch { }
				}

				// if we haven't succeeded, try the WinUSB driver
				if (linkPort == null) {
					linkPort = new SilverLinkWinUsb();
					if (!((SilverLinkWinUsb)linkPort).IsOpen) {
						Console.WriteLine("Could not open the USB SilverLink.");
						linkPort = null;
						return;
					}
				}

			} else {
				// try to open the serial cable port
				SerialPort serialPort;
				try {
					serialPort = new SerialPort(cablePortName);
					serialPort.Open();
				} catch (Exception ex) {
					Console.WriteLine("Could not open the serial port: {0}", ex.Message);
					return;
				}

				if (serialPort != null) {
					linkPort = new BlackLink(serialPort);
				} else {
					Console.WriteLine("No serial port to access.");
					return;
				}
			}

			// final check
			if (linkPort == null) {
                Console.WriteLine("Link cable port not opened.");
                return;
			}

			// get ready to go!
			try {

				retry:;

				Console.WriteLine();
				Console.WriteLine("Sending '{0}' ({1} bytes).", payloadName, payload.Length);
				Console.WriteLine("Flash will be {0}.", unlockFlash ? "unlocked" : "locked");
				if (fixedPacketSize) {
					Console.WriteLine("Fixed packet size with a packet size counter on stack at ${0:X2}.", fixedPacketSizeStackPointer);
				} else if (zeroPadPacket) {
					if (zeroPadPacketCount < 0) {
						Console.WriteLine("Packet zero-padded until link activity fails.");
					} else {
						Console.WriteLine("Packet zero-padded with {0} bytes.", zeroPadPacketCount);
					}
				}
				var portName = linkPort.GetType().GetCustomAttribute<DescriptionAttribute>()?.Description;
				if (string.IsNullOrEmpty(portName)) portName = linkPort.GetType().Name;
				Console.WriteLine("Using {0} on {1}.", portName, cablePortName);
				Console.WriteLine();

				Console.WriteLine("Ensure the calculator is at the 'Waiting... Please install calculator software now.' screen.");
				Console.Write("Press any key when ready, or Escape to cancel...");
				if (Console.ReadKey(true).Key == ConsoleKey.Escape) {
					Console.WriteLine("Cancel");
					return;
				} else {
					Console.WriteLine("OK");
				}

				if (unlockFlash) {

					Console.Write("Unlocking flash...");
					if (!linkPort.TryWriteBytes(new byte[] { 0x23, 0xC9, 0x00, 0x01 })) {
						Console.WriteLine("Failed!");
						return;
					}

					byte[] flashUnlock = new byte[256];
					flashUnlock[0] = 0xFF;
					flashUnlock[1] = 0xFF;
					flashUnlock[2] = 0xFF;
					flashUnlock[3] = 0xFF;
					flashUnlock[4] = 0x25;

					if (!linkPort.TryWriteBytes(flashUnlock)) {
						Console.WriteLine("Failed!");
						return;
					}

					ushort flashUnlockChecksum = 0;
					foreach (var b in flashUnlock) {
						flashUnlockChecksum += b;
					}

					if (!linkPort.TryWriteBytes(new byte[] { (byte)(flashUnlockChecksum >> 0), (byte)(flashUnlockChecksum >> 8) })) {
						Console.WriteLine("Failed!");
						return;
					}
					Console.WriteLine("OK");

					Console.Write("Getting ack...");
					{
						DateTime timeout = DateTime.Now;
						timeout += TimeSpan.FromSeconds(2);
						while (!linkPort.TryReadBytes(out byte[] b, 4)) {
							if (DateTime.Now > timeout) {
								Console.WriteLine("Failed!");
								return;
							}
						}
					}
					Console.WriteLine("OK");
				}

				ushort largePacketSize = (ushort)(fixedPacketSize ? 0xFFFF : (0x10000 - 0x867D));


				Console.Write("Sending large packet header...");
				if (!linkPort.TryWriteBytes(new byte[] { 0x23, 0x06, (byte)largePacketSize, (byte)(largePacketSize >> 8) })) {
					Console.WriteLine("Failed!");
					return;
				}
				Console.WriteLine("OK");

				// clear checksum
				ushort checksum = 0;

				Console.Write("Advancing pointer to 9090h...");
				for (int i = 0x867D; i < 0x9090; ++i) {
					if (!linkPort.TryWriteByte(0x00)) {
						Console.WriteLine("Failed!");
						return;
					}
				}
				Console.WriteLine("OK");

				Console.Write("Writing jump instruction...");
				if (!linkPort.TryWriteBytes(new byte[] { 0xC3, 0x95, 0x9D })) { // JP 9D95h
					Console.WriteLine("Failed!");
					return;
				}
				Console.WriteLine("OK");

				// offset checksum
				checksum += 0xC3;
				checksum += 0x95;
				checksum += 0x9D;

				Console.Write("Advancing pointer to 9D95h...");
				for (int i = 0x9093; i < 0x9D95; ++i) {
					if (!linkPort.TryWriteByte(0x00)) {
						Console.WriteLine("Failed!");
						return;
					}
				}
				Console.WriteLine("OK");

				Console.Write("Sending payload...");
				if (!linkPort.TryWriteBytes(payload)) {
					Console.WriteLine("Failed!");
					return;
				}
				Console.WriteLine("OK");

				// recalculate checksum
				foreach (var b in payload) {
					checksum += b;
				}

				if (fixedPacketSize) {

					Console.Write("Advancing to stack...");
					for (int i = 0x9D95 + payload.Length; i < fixedPacketSizeStackPointer; ++i) {
						if (!linkPort.TryWriteByte(0x00)) {
							Console.WriteLine("Failed!");
							return;
						}
					}
					Console.WriteLine("OK");

					Console.Write("Writing data length...");
					ushort dataLength = (ushort)(0x10000 - fixedPacketSizeStackPointer);
					if (!linkPort.TryWriteByte((byte)dataLength)) {
						Console.WriteLine("Failed!");
						return;
					}

					checksum += (byte)dataLength;
					dataLength >>= 8;

					if (!linkPort.TryWriteByte((byte)dataLength)) {
						Console.WriteLine("Failed!");
						return;
					}

					checksum += (byte)dataLength;

					Console.WriteLine("OK");

					Console.Write("Filling stack with 90h...");
					for (int i = fixedPacketSizeStackPointer + 2; i < 0x10000; ++i) {
						if (!linkPort.TryWriteByte(0x90)) {
							Console.WriteLine("Failed!");
							return;
						}
						checksum += 0x90;
					}
					Console.WriteLine("OK");

					Console.Write("Sending checksum...");
					if (!linkPort.TryWriteByte((byte)(checksum >> 0))) {
						Console.WriteLine("Failed!");
						return;
					}
					if (!linkPort.TryWriteByte((byte)(checksum >> 8))) {
						Console.WriteLine("Failed!");
						return;
					}
					Console.WriteLine("OK");

				} else {

					Console.Write("Filling stack with 90h...");
					for (int i = 0x9D95 + payload.Length; i < 0x10000; ++i) {
						if (!linkPort.TryWriteByte(0x90)) {
							Console.WriteLine("Failed!");
							return;
						}
						checksum += 0x90;
					}
					Console.WriteLine("OK");

					Console.Write("Adjusting checksum...");
					ushort checksumCorrection = (ushort)(-checksum);

					while (checksumCorrection > 0xFF) {
						if (!linkPort.TryWriteByte((byte)0xFF)) {
							Console.WriteLine("Failed!");
							return;
						}
						checksumCorrection -= 0xFF;
					}
					while (checksumCorrection > 0) {
						if (!linkPort.TryWriteByte((byte)checksumCorrection)) {
							Console.WriteLine("Failed!");
							return;
						}
						checksumCorrection -= (byte)checksumCorrection;
					}

					Console.WriteLine("OK");

					if (zeroPadPacketCount >= 0) {
						Console.Write("Sending zeroes...");
						for (int i = 0; i < zeroPadPacketCount; ++i) {
							if (!linkPort.TryWriteByte(0x00)) {
								Console.WriteLine("Failed!");
								return;
							}
						}
						Console.WriteLine("OK");


					} else {
						int zeroesWritten = 0;
						Console.Write("Counting zeroes...");
						while (linkPort.TryWriteByte(0x00)) {
							++zeroesWritten;
						}
						Console.WriteLine("OK ({0})", zeroesWritten);
						zeroPadPacketCount = zeroesWritten;
						Console.WriteLine("Retrying now that the number of zeroes has been counted.");
						goto retry;
					}

				}

				Console.Write("Getting ack...");
				{
					DateTime timeout = DateTime.Now;
					timeout += TimeSpan.FromSeconds(2);
					while (!linkPort.TryReadBytes(out byte[] b, 4)) {
						if (DateTime.Now > timeout) {
							Console.WriteLine("Failed!");
							return;
						}
					}
				}
				Console.WriteLine("OK");

				Console.WriteLine();
				Console.WriteLine("Done");

			} finally {
				if (linkPort is IDisposable) {
					((IDisposable)linkPort).Dispose();
				}
			}
		}

	}
}
