﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Management;
using System.Runtime.InteropServices;

namespace TI83PlusBootExec {

	[Description("Silver Link (TI driver)")]
	internal class SilverLinkTI : LinkCable, IDisposable {

		#region Win32 P/Invoke

		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		public static extern SafeFileHandle CreateFile(
			[In, MarshalAs(UnmanagedType.LPTStr)] string lpFileName,
			[In, MarshalAs(UnmanagedType.U4)] FileAccess dwDesiredAccess,
			[In, MarshalAs(UnmanagedType.U4)] FileShare dwShareMode,
			[In, Optional] IntPtr lpSecurityAttributes,
			[In, MarshalAs(UnmanagedType.U4)] FileMode dwCreationDisposition,
			[In, MarshalAs(UnmanagedType.U4)] FileAttributes dwFlagsAndAttributes,
			[In, Optional] IntPtr hTemplateFile
		);

		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		static extern bool ReadFile(
			[In] SafeFileHandle hFile,
			[Out] byte[] lpBuffer,
			[In] uint nNumberOfBytesToRead,
			[Out, Optional] out uint lpNumberOfBytesRead,
			[In, Out, Optional] IntPtr lpOverlapped
		);

		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		static extern bool WriteFile(
			[In] SafeFileHandle hFile,
			[In] byte[] lpBuffer,
			[In] uint nNumberOfBytesToWrite,
			[Out, Optional] out uint lpNumberOfBytesWritten,
			[In, Out, Optional] IntPtr lpOverlapped
		);

		#endregion

		#region Cable enumeration

		/// <summary>
		/// Finds the Silver Link cables connected to the computer.
		/// </summary>
		/// <returns>An array of available Silver Link cables.</returns>
		public static SilverLinkTI[] GetCables() {
			List<SilverLinkTI> cables = new List<SilverLinkTI>();

			// Look for PnP devices with the matching USB VID/PID and that are using the SilvrLnk service.
			using (var cableSearcher = new ManagementObjectSearcher(@"SELECT * FROM Win32_PnPEntity WHERE Service = 'SilvrLnk' AND DeviceID LIKE 'USB\\VID_0451&PID_E001%'")) {
				foreach (var cableDevice in cableSearcher.Get()) {
					// Device path is device ID with \ replaced with # and with the GUID_DEVINTERFACE_USB_DEVICE on the end.
					cables.Add(new SilverLinkTI(@"\\?\" + ((string)cableDevice["DeviceID"]).Replace('\\', '#') + "#{a5dcbf10-6530-11d2-901f-00c04fb951ed}"));
					cableDevice.Dispose();
				}
			}
			return cables.ToArray();
		}

		#endregion

		#region Opening/Closing

		private SafeFileHandle handle;
		private readonly string devicePath;

		/// <summary>
		/// Creates an instance of the Silver Link cable class for a particular USB device path.
		/// </summary>
		/// <param name="devicePath">The USB device path to create the instance for.</param>
		public SilverLinkTI(string devicePath) {
			this.devicePath = devicePath;
		}

		/// <summary>
		/// Opens the Silver Link cable.
		/// </summary>
		public void Open() {
			if (handle != null) {
				throw new InvalidOperationException("Device is already open.");
			}
			var handleAttempt = CreateFile(devicePath, FileAccess.ReadWrite, FileShare.None, IntPtr.Zero, FileMode.Open, FileAttributes.Device, IntPtr.Zero);

			if (handleAttempt.IsInvalid) {
				handleAttempt.Dispose();
				throw new InvalidOperationException("Could not open device.");
			} else {
				handle = handleAttempt;
			}
		}

		/// <summary>
		/// Closes the Silver Link cable.
		/// </summary>
		public void Close() {
			if (handle == null) {
				throw new InvalidOperationException("Device is already closed.");
			}
			handle.Close();
			handle.Dispose();
			handle = null;
		}

		/// <summary>
		/// Gets a value indicating whether the Silver Link is closed.
		/// </summary>
		public bool IsClosed {
			get { return handle == null || handle.IsClosed; }
		}

		#endregion

		#region Reading/Writing

		/// <summary>
		/// Reads data from the Silver Link.
		/// </summary>
		/// <param name="buffer">The buffer to read into.</param>
		/// <param name="count">The maximum number of bytes to read.</param>
		/// <returns>The number of bytes that were read.</returns>
		public int Read(byte[] buffer, int count) {
			if (IsClosed) throw new InvalidOperationException("Device is closed.");
			if (!ReadFile(handle, buffer, (uint)count, out uint bytesRead)) throw new InvalidOperationException("Could not read from device.");
			return (int)bytesRead;
		}
		
		public override bool TryReadByte(out byte value) {
			try {
				var buffer = new byte[1];
				if (Read(buffer, 1) == 1) {
					value = buffer[0];
					return true;
				} else {
					value = default;
					return false;
				}
			} catch {
				value = default;
				return false;
			}
		}

		public override bool TryReadBytes(out byte[] values, int count) {
			try {
				values = new byte[count];
				return Read(values, count) == count;
			} catch {
				values = null;
				return false;
			}
		}

		/// <summary>
		/// Writes data to the Silver Link.
		/// </summary>
		/// <param name="buffer">The buffer to write from.</param>
		/// <param name="count">The maximum number of bytes to write.</param>
		/// <returns>The number of bytes that were written.</returns>
		public int Write(byte[] buffer, int count) {
			if (IsClosed) throw new InvalidOperationException("Device is closed.");
			if (!WriteFile(handle, buffer, (uint)count, out uint bytesWritten)) throw new InvalidOperationException("Could not write to device.");
			return (int)bytesWritten;
		}

		public override bool TryWriteByte(byte value) {
			try {
				var buffer = new byte[] { value };
				return Write(buffer, 1) == 1;
			} catch {
				return false;
			}
		}

		public override bool TryWriteBytes(byte[] values) {
			try {
				return Write(values, values.Length) == values.Length;
			} catch {
				return false;
			}
		}

		#endregion

		#region IDisposable

		private bool disposedValue;

		protected virtual void Dispose(bool disposing) {
			if (!disposedValue) {

				if (disposing) {
					if (handle != null) {
						handle.Dispose();
						handle = null;
					}
				}

				handle = null;
				disposedValue = true;
			}
		}

		public void Dispose() {
			Dispose(disposing: true);
			GC.SuppressFinalize(this);
		}

		#endregion


	}

}
